#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266HTTPClient.h>
#include <ESP8266mDNS.h>
#include <ArduinoOTA.h>
#include <U8g2lib.h>
#include <EEPROM.h>
#include <TimeLib.h>


#define gpio_SDA    4   // I2C Data pin
#define gpio_SCL    5   // I2C Clock pin
#define gpio_Pulse1 12  // Pulse In OC1
#define gpio_Pulse2 13  // Pulse In OC2
#define OTAPASSW    "password"  // Password used for OTA update
#define avgInterval 60  // Seconds on average interval
#define memSize    512  // Number of bytes for EEPROM
#define TZoneOffset 60  // Minutes offset from UTC, CET = UTC + 60 mins
#define TZoneDST    60  // Minutes offset for Daylight Saving Time  (i.e. 'S' after time datagram)

char espHostname[30];

char ssid[] = "YourSSID";
char password[] = "YourWifiPassword";

long mEVLT = 0; // consumption low tariff (0,001kWh)
long mEVHT = 0; // consumption high tariff (0,001kWh)
long mEPLT = 0; // production low tariff (0,001kWh)
long mEPHT = 0; // production high tariff (0,001kWh)
long mEAV = 0;  // actual consumption (0,001kW)
long mEAP = 0;  // actual production (0,001kW)
long mCT = 0;   // actual tariff (1/2)
long mGVT = 0;  // m-bus reading gas (0,001m3)
long mWVT = 0;  // m-bus reading water (0,001m3)
long oEVLT = 0; // consumption low tariff (0,001kWh)
long oEVHT = 0; // consumption high tariff (0,001kWh)
long oEPLT = 0; // production low tariff (0,001kWh)
long oEPHT = 0; // production high tariff (0,001kWh)
long oGVT = 0;  // m-bus reading gas (0,001m3)
long oWVT = 0;  // m-bus reading water (0,001m3)

unsigned long pWVT1 = 0; // Puls sensor 1 water
unsigned long pWVT2 = 0; // Puls sensor 2 water

char dsptime[24];
boolean newData = false;

long avgPwr[avgInterval];  // calculating average
long avgCnt = 0;  // items counter
long mi, av, ma, mMIN, mAVG, mMAX;
boolean hasAVG = false;
unsigned long avgTimer;

unsigned long tmDelay, tmEEPROM;

int wRSSI = 0;  // The RSSI indicator, scaled from 0 (worse) to 4 (best)

String datagram = "";
String debugme = "";
String httpdebug = "";
String chipid = "";

ESP8266WebServer server(80);
// Use this line for 0.96" OLED Display
U8G2_SSD1306_128X64_NONAME_F_SW_I2C u8g2(U8G2_R0, gpio_SCL, gpio_SDA, U8X8_PIN_NONE);   // All Boards without Reset of the Display
// Use this line for 1.3" OLED Display
// U8G2_SH1106_128X64_NONAME_F_SW_I2C u8g2(U8G2_R0, gpio_SCL, gpio_SDA, U8X8_PIN_NONE);   // All Boards without Reset of the Display

time_t tmLocal, tmUTC;
String tmpTime;


void setup(void) {
  pinMode(gpio_Pulse1, INPUT);
  pinMode(gpio_Pulse2, INPUT);
  
  sprintf(espHostname, "ESP%X", ESP.getChipId());

  u8g2.setI2CAddress(0x78);
  u8g2.begin();
  twi_setClock(400000);
  
  Serial.begin(115200);
  WiFi.hostname(espHostname);
  WiFi.begin(ssid, password);
  Serial.println("");

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("");
  Serial.print("Connected to ");
  Serial.println(ssid);
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());

  if (MDNS.begin(espHostname)) {
    Serial.println("MDNS responder started");
  }

  server.on("/", handleXML);
  server.on("/debug", handleDebug);
  server.on("/reset", handleReset);
  server.on("/offset", handleOffset);
  server.onNotFound(handleXML);

  server.begin();
  Serial.println("HTTP server started");


  setupOTA();

  tmDelay = millis();
  tmEEPROM = tmDelay;
  avgTimer = tmDelay;

  EEPROM.begin(memSize);
  getOffsetCounters();
  
  attachInterrupt(digitalPinToInterrupt(gpio_Pulse1), irqSensorPulse1, FALLING);
  attachInterrupt(digitalPinToInterrupt(gpio_Pulse2), irqSensorPulse2, FALLING);  
}


void loop(void) {
  server.handleClient();
  ArduinoOTA.handle();
  
  if(millis() - tmDelay >= 1000) {
    updateDisplay();
    tmDelay = millis();    
  }

  if(millis() - tmEEPROM >= 300000) {
    tmEEPROM = millis();
    saveOffsetCounters();
  }
  
  processAverage();
  
  if(Serial.available()) {
    char c = (char) Serial.read() & 127;
    if(c == '/') {
      datagram = "";
      newData = false;
    }
    datagram += String((char) c);
    if(c == '!') {
      // Process datagram
      decodeDatagram();
      updateDatabase();
    }
  }
}



void handleDebug() {
  String message = "Server up and running\n\r";
  message += datagram;
  message += "\n\r\n\r-------- Debug Data ----------\n\r";
  message += debugme;
  message += "\n\r\n\r-------- Network Info --------\n\r";
  message += "RSSI: ";
  message += String(WiFi.RSSI());
  message += "dBm (Indicator ";
  message += String(wRSSI);
  message += ")\n\rChip info: 0x";
  message += String(ESP.getChipId(), HEX);
  message += "\n\rMAC Addr: ";
  message += String(WiFi.macAddress());
  message += "\n\rTime: ";
  message += tmpTime;
  message += "\n\rUTC Time: ";
  message += String(tmUTC);
  message += "\n\r\n\r-------- HTTP Data -----------\n\r";
  message += httpdebug;
  server.send(200, "text/plain", message);
}


void handleReset() {
  String message = "Resetting in 5 seconds...\n\r";
  server.send(200, "text/plain", message);
  saveOffsetCounters();
  delay(5000);
  ESP.restart();
}


void handleOffset() {
  if(server.arg("oc1") != "") {
    pWVT1 = (unsigned long) server.arg("oc1").toInt();
  }
  if(server.arg("oc2") != "") {
    pWVT2 = (long) server.arg("oc2").toInt();
  }
  saveOffsetCounters(); // Write new values to EEPROM
  server.send(200, "text/plain", "Offset processed");
}



void handleXML() {
  String message = "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n\r";
  message += "<P1Log>\n\r";

  message += "  <pwrConsT1>";
  message += String(mEVLT);
  message += "</pwrConsT1>\n\r";
  message += "  <pwrConsT2>";
  message += String(mEVHT);
  message += "</pwrConsT2>\n\r";
  message += "  <pwrProdT1>";
  message += String(mEPLT);
  message += "</pwrProdT1>\n\r";
  message += "  <pwrProdT2>";
  message += String(mEPHT);
  message += "</pwrProdT2>\n\r";

  message += "  <pwrActCons>";
  message += String(mEAV);
  message += "</pwrActCons>\n\r";
  message += "  <pwrActProd>";
  message += String(mEAP);
  message += "</pwrActProd>\n\r";
  message += "  <pwrTariff>";

  message += String(mCT);
  message += "</pwrTariff>\n\r";

  message += "  <gasConsT>";
  message += String(mGVT);
  message += "</gasConsT>\n\r";
  message += "  <wtrConsT>";
  message += String(mWVT);
  message += "</wtrConsT>\n\r";
  message += "  <ctrPulse1>";
  message += String(pWVT1);
  message += "</ctrPulse1>\n\r";
  message += "  <ctrPulse2>";
  message += String(pWVT2);
  message += "</ctrPulse2>\n\r";
  
  message += "  <intPwrMin>";
  message += String(mMIN);
  message += "</intPwrMin>\n\r";
  message += "  <intPwrAvg>";
  message += String(mAVG);
  message += "</intPwrAvg>\n\r";
  message += "  <intPwrMax>";
  message += String(mMAX);
  message += "</intPwrMax>\n\r";

  message += "</P1Log>\n\r";
  
  server.send(200, "text/xml", message);
}


void setupOTA() {
  // Port defaults to 8266
  ArduinoOTA.setPort(8266);
  // Hostname defaults to ESP[ChipID]
  ArduinoOTA.setHostname(espHostname);
  ArduinoOTA.setPassword(OTAPASSW);
  ArduinoOTA.onStart([]() {
    Serial.println("Start");
    saveOffsetCounters();
  });
  ArduinoOTA.onEnd([]() {
    Serial.println("\nEnd");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
    else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
    else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
    else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
    else if (error == OTA_END_ERROR) Serial.println("End Failed");
  });
  ArduinoOTA.begin();
}


void checkRSSI() {
  int x = WiFi.RSSI();
  wRSSI = 0;
  if(x < -30) { // -35dBm is over saturated signal
    if(x > -55) {
      wRSSI = 4; // Best quality connection
    } else if(x > -65) {
      wRSSI = 3; // Usable connection
    } else if(x > -72) {
      wRSSI = 2; // Poor connection, high latencies and drops
    } else if(x > -80) {
      wRSSI = 1; // Worst connection, better not use this one
    }
  }
}


void updateDisplay() {
  u8g2.clearBuffer();
  u8g2.setFont(u8g2_font_victoriabold8_8r);
  u8g2.setFontPosTop();
  u8g2.setDrawColor(1);
  u8g2.drawBox(0, 0, u8g2.getStrWidth("P1 LOGGER") + 4, 10);
  u8g2.setDrawColor(0);
  u8g2.drawStr(1, 0, "P1 LOGGER");
  u8g2.setDrawColor(1);
  int b = (millis() - avgTimer) / 2000;
  if(b > 30) {
    b = 30;
  }
  u8g2.drawFrame(122, 29, 5, 32);
  u8g2.drawBox(123, 60 - b, 3, b + 1);
  if(millis()/1000 % 4 < 2) {
    dtostrf((float) mEVLT / 1000.0, 10, 3, dsptime);
    u8g2.drawStr(118 - u8g2.getStrWidth(dsptime), 17, dsptime);
    dtostrf((float) mEVHT / 1000.0, 10, 3, dsptime);  
    u8g2.drawStr(118 - u8g2.getStrWidth(dsptime), 29, dsptime);
    dtostrf((float) mGVT / 1000.0, 10, 3, dsptime);  
    u8g2.drawStr(118 - u8g2.getStrWidth(dsptime), 41, dsptime);
    dtostrf((float) mEAV / 1000.0, 10, 3, dsptime);
    u8g2.drawStr(118 - u8g2.getStrWidth(dsptime), 53, dsptime);
    u8g2.setFont(u8g2_font_victoriamedium8_8r);
    u8g2.drawStr(0, 17, "T1:");
    u8g2.drawStr(0, 29, "T2:");
    u8g2.drawStr(0, 41, "Gas:");
    u8g2.drawStr(0, 53, "Act:");
  } else {
    dtostrf((float) mEPLT / 1000.0, 10, 3, dsptime);
    u8g2.drawStr(118 - u8g2.getStrWidth(dsptime), 17, dsptime);
    dtostrf((float) mEPHT / 1000.0, 10, 3, dsptime);  
    u8g2.drawStr(118 - u8g2.getStrWidth(dsptime), 29, dsptime);
    sprintf(dsptime, "%6d", pWVT1);
    u8g2.drawStr(118 - u8g2.getStrWidth(dsptime), 42, dsptime);
    sprintf(dsptime, "%6d", pWVT2);
    u8g2.drawStr(118 - u8g2.getStrWidth(dsptime), 53, dsptime);
    u8g2.setFont(u8g2_font_victoriamedium8_8r);
    u8g2.drawStr(0, 17, "P1:");
    u8g2.drawStr(0, 29, "P2:");
    u8g2.drawStr(0, 41, "OC1:");
    u8g2.drawStr(0, 53, "OC2:");
  }
  
  if(WiFi.status() == WL_CONNECTED) {
    checkRSSI();
    for(int i = 1; i < 5; i++) {
      if(wRSSI >= i) {
        u8g2.drawBox(110 + i * 3, 9 - i * 2, 3, i * 2);
      } else {
        u8g2.drawFrame(110 + i * 3, 9 - i * 2, 4, i * 2);
      }
    } 
  } else { // No wifi connection available
        u8g2.drawLine (110, 1, 119, 10);
        u8g2.drawLine (111, 0, 120, 9);
        u8g2.drawLine (110, 9, 119, 0);
        u8g2.drawLine (111, 10, 120, 1);
  }

  u8g2.sendBuffer();
}

void decodeDatagram() {
  int x, y;
  String t;
  char inChar;
  long tmpVal;
  
  // time from meter
  x = datagram.indexOf("0-0:1.0.0(");
  if(x >= 0) {
    y = datagram.indexOf(")", x);
    if(y > 0 && y < x + 24) {
      t = datagram.substring(x + 10, y);
      tmpTime = t.substring(6, 8) + ":" + t.substring(8, 10) + ":" + t.substring(10, 12) + "   " + t.substring(4, 6) + "-" + t.substring(2, 4) + "-20" + t.substring(0, 2);
      setTime(t.substring(6, 8).toInt(), t.substring(8, 10).toInt(), t.substring(10, 12).toInt(), t.substring(4, 6).toInt(), t.substring(2, 4).toInt(), t.substring(0, 2).toInt() + 2000);
      tmLocal = now();
      if(t.indexOf("S") > 0) {
        tmUTC = tmLocal - 60 * (TZoneOffset + TZoneDST);
      } else {
        tmUTC = tmLocal - 60 * TZoneOffset;
      }
      if(tmUTC % 604800 == 99000)   // Reboot every Monday at 03:30 UTC
      {
        handleReset();
      }
    }
  }

  // consumption low tariff
  x = datagram.indexOf("1-0:1.8.1(");
  if(x >= 0) {
    y = datagram.indexOf("*kWh)", x);
    if(y > 0 && y < x + 22) {
      t = datagram.substring(x + 10, y);
      tmpVal = (long) (t.toFloat() * 1000.0);
      if(tmpVal >= oEVLT && tmpVal >= 0) {
        oEVLT = mEVLT;
        mEVLT = tmpVal;
        newData = true;
      }
    }
  }

  // consumption high tariff
  x = datagram.indexOf("1-0:1.8.2(");
  if(x >= 0) {
    y = datagram.indexOf("*kWh)", x);
    if(y > 0 && y < x + 22) {
      t = datagram.substring(x + 10, y);
      tmpVal = (long) (t.toFloat() * 1000.0);
      if(tmpVal >= oEVHT && tmpVal >= 0) {
        oEVHT = mEVHT;
        mEVHT = tmpVal;
        newData = true;
      }
    }
  }

  // production low tariff
  x = datagram.indexOf("1-0:2.8.1(");
  if(x >= 0) {
    y = datagram.indexOf("*kWh)", x);
    if(y > 0 && y < x + 22) {
      t = datagram.substring(x + 10, y);
      tmpVal = (long) (t.toFloat() * 1000.0);
      if(tmpVal >= oEPLT && tmpVal >= 0) {
        oEPLT = mEPLT;
        mEPLT = tmpVal;
        newData = true;
      }
    }
  }

  // production high tariff
  x = datagram.indexOf("1-0:2.8.2(");
  if(x >= 0) {
    y = datagram.indexOf("*kWh)", x);
    if(y > 0 && y < x + 22) {
      t = datagram.substring(x + 10, y);
      tmpVal = (long) (t.toFloat() * 1000.0);
      if(tmpVal >= oEPHT && tmpVal >= 0) {
        oEPHT = mEPHT;
        mEPHT = tmpVal;
        newData = true;
      }
    }
  }

  // actual consumption
  x = datagram.indexOf("1-0:1.7.0(");
  if(x >= 0) {
    y = datagram.indexOf("*kW)", x);
    if(y > 0 && y < x + 18) {
      t = datagram.substring(x + 10, y);
      tmpVal = (long) (t.toFloat() * 1000.0);
      if(tmpVal >= 0) {
        mEAV = tmpVal;
        newData = true;
      }
    }
  }

  // actual production
  x = datagram.indexOf("1-0:2.7.0(");
  if(x >= 0) {
    y = datagram.indexOf("*kW)", x);
    if(y > 0 && y < x + 18) {
      t = datagram.substring(x + 10, y);
      tmpVal = (long) (t.toFloat() * 1000.0);
      if(tmpVal >= 0) {
        mEAP = tmpVal;
        newData = true;
      }
    }
  }

  // actual tariff
  x = datagram.indexOf("0-0:96.14.0(");
  if(x >= 0) {
    y = datagram.indexOf(")", x);
    if(y > 0 && y < x + 18) {
      t = datagram.substring(x + 12, y);
      mCT = (long) (t.toInt());
      newData = true;
    }
  }
  
  // actual gas consumption
  x = datagram.indexOf("0-1:24.2.1(");
  if(x >= 0) {
    x = datagram.indexOf(")(", x);
    y = datagram.indexOf("*m3)", x);
    if(y > 0 && y < x + 13) {
      t = datagram.substring(x + 2, y);
      tmpVal = (long) (t.toFloat() * 1000.0);
      if(tmpVal >= oGVT && tmpVal >= 0) {
        oGVT = mGVT;
        mGVT = tmpVal;
        newData = true;
      }
    }
  }

  // actual water consumption
  x = datagram.indexOf("0-2:96.1.0(");
  if(x >= 0) {
    x = datagram.indexOf(")(", x);
    y = datagram.indexOf("*m3)", x);
    if(y > 0 && y < x + 13) {
      t = datagram.substring(x + 2, y);
      tmpVal = (long) (t.toFloat() * 1000.0);
      if(tmpVal >= oWVT && tmpVal >= 0) {
        oWVT = mWVT;
        mWVT = tmpVal;
        newData = true;
      }
    }
  }
  mWVT = pWVT1; // Fix the water using OC1

  avgPwr[avgCnt] = mEAV - mEAP;
  avgCnt++;
  if(avgCnt > avgInterval - 1) {
    avgCnt = avgInterval - 1;
  }
}

void processAverage() {
  // build an average every 60 seconds
  if(millis() - avgTimer >= avgInterval * 1000) {
    avgTimer = millis();
    av = 0;
    mi = 2147483646L;
    ma = -2147483646L;
    for(int i=0; i < avgCnt; i++) {
      av = av + avgPwr[i];
      if(avgPwr[i] > ma) {
        ma = avgPwr[i];
      }
      if(avgPwr[i] < mi) {
        mi = avgPwr[i];
      }
      avgPwr[i] = 0;
    }
    if(avgCnt > 0) {
      av = av / avgCnt;
    } else {
      av = 0;
    }
    avgCnt = 0;
    mMIN = mi;
    mAVG = av;
    mMAX = ma;
    hasAVG = true;    
  }
}

void updateDatabase() {
  
  String myURL = "http://yourserver.url/log.php?";
  myURL += "mEVLT=";
  myURL += String(mEVLT);
  myURL += "&mEVHT=";
  myURL += String(mEVHT);
  myURL += "&mEPLT=";
  myURL += String(mEPLT);
  myURL += "&mEPHT=";
  myURL += String(mEPHT);
  myURL += "&mEAV=";
  myURL += String(mEAV);
  myURL += "&mEAP=";
  myURL += String(mEAP);
  myURL += "&mCT=";
  myURL += String(mCT);
  myURL += "&mGVT=";
  myURL += String(mGVT);
  myURL += "&mWVT=";
  myURL += String(mWVT);
  myURL += "&mOC1=";
  myURL += String(pWVT1);
  myURL += "&mOC2=";
  myURL += String(pWVT2);
  if(hasAVG) {
    myURL += "&mMIN=";
    myURL += String(mMIN);
    myURL += "&mAVG=";
    myURL += String(mAVG);
    myURL += "&mMAX=";
    myURL += String(mMAX);
    hasAVG = false;
  }
  debugme = myURL;
  debugme += "\n\rSend state: ";
  debugme += String(newData);
  
  if(WiFi.status() == WL_CONNECTED && newData) {
    HTTPClient http;
    http.begin(myURL);
    int httpCode = http.GET();
    if(httpCode > 0) {
      // HTTP header has been send and Server response header has been handled
      
      // file found at server
      if(httpCode == HTTP_CODE_OK) {
        String payload = http.getString();
        httpdebug = payload;
      }
    } else {
      // Something failed
    }
    http.end();
  }
}

void getOffsetCounters() {
  pWVT1 = (EEPROM.read(0) << 24) + (EEPROM.read(1) << 16) + (EEPROM.read(2) << 8) + EEPROM.read(3);
  pWVT2 = (EEPROM.read(4) << 24) + (EEPROM.read(5) << 16) + (EEPROM.read(6) << 8) + EEPROM.read(7);
}

void saveOffsetCounters() {
  EEPROM.write(0, (pWVT1 >> 24) & 0xFF);
  EEPROM.write(1, (pWVT1 >> 16) & 0xFF);
  EEPROM.write(2, (pWVT1 >> 8) & 0xFF);
  EEPROM.write(3, pWVT1 & 0xFF);
  EEPROM.write(4, (pWVT2 >> 24) & 0xFF);
  EEPROM.write(5, (pWVT2 >> 16) & 0xFF);
  EEPROM.write(6, (pWVT2 >> 8) & 0xFF);
  EEPROM.write(7, pWVT2 & 0xFF);
  EEPROM.commit();    // Write EEPROM data to flash
}

void irqSensorPulse1() {
  pWVT1++;
}

void irqSensorPulse2() {
  pWVT2++;
}

